package dammfine.jife;

import java.util.Iterator;
import java.util.List;
import java.util.LinkedList;
import java.util.Properties;
import java.net.Socket;
import java.io.IOException;

import dammfine.dict.*;
import dammfine.dict.receiving.*;
import dammfine.dict.sending.*;
public class DictSource implements DefinitionSource
{
	
	public String server;
	public int port;
	
	public void init( String name, Properties props )
	{
		this.server = props.getProperty( name + ".server" );
		this.port = Integer.parseInt( props.getProperty( name + ".port" ) );
	}
	
	public List getDictionaries()
		throws JifeException 
	{
		try
		{
			Socket socket = new Socket( server, port );
			try
			{
				DictConnection conn = new DictConnection( socket );
				
				
				conn.sendCommand( new ClientCommand() );
				List results = conn.sendCommand( new ShowDatabasesCommand() );
				conn.sendCommand( new QuitCommand() );
				
				// In the results we will find a
				// DatabasesPresentCommand, which should give us
				// what we want.
				Iterator resultsI = results.iterator();
				while( resultsI.hasNext() )
				{
					Object o = resultsI.next();
					if( o instanceof DatabasesPresentCommand )
					{
						DatabasesPresentCommand dpc = (DatabasesPresentCommand)o;
						return dpc.getDatabases();
					}
				}
				
				// No results in result set
				// We should look at the commands returned and find
				// out if there is a reason other than the word wasnt
				// found (like, invalid dictionary).
				return null;
			}
			finally
			{
				socket.close();
			}
		}
		catch( IOException ioe )
		{
			throw new JifeException( ioe.getMessage() );
		}
	}
	
	public List getDefinition( String word )
		throws JifeException 
	{
		return getDefinition( "wn", word );
	}

	public List getDefinition( String dict, String word )
		throws JifeException 
	{
		try
		{
			Socket socket = new Socket( server, port );
			try
			{
				DictConnection conn = new DictConnection( socket );


				conn.sendCommand( new ClientCommand() );
				List results = conn.sendCommand( new DefineCommand( dict, word ) );
				conn.sendCommand( new QuitCommand() );
				
				// List of strings, our result.
				List resultLines = new LinkedList();
				 
				// In the results we will find a
				// DefinitionCommand, which should give us
				// what we want.
				Iterator resultsI = results.iterator();
				boolean found = false;
				while( resultsI.hasNext() )
				{
					Object o = resultsI.next();
					if( o instanceof DefinitionCommand )
					{
						found = true;
						DefinitionCommand dpc = (DefinitionCommand)o;
						resultLines.addAll( dpc.getDefinitions() );
					}
				}
				
				if( ! found )
				{
					// No results in result set
					// We should look at the commands returned and find
					// out if there is a reason other than the word wasnt
					// found (like, invalid dictionary).
					return null;
				}
				else
				{
					return resultLines;
				}
			}
			finally
			{
				socket.close();
			}
		}
		catch( IOException ioe )
		{
			throw new JifeException( ioe.getMessage() );
		}
	}
}

