package dammfine.jife;

import f00f.net.irc.martyr.IRCConnection;
import f00f.net.irc.martyr.AutoRegister;
import f00f.net.irc.martyr.AutoResponder;
import f00f.net.irc.martyr.AutoReconnect;
import f00f.net.irc.martyr.AutoJoin;
import f00f.net.irc.martyr.GenericAutoService;
import f00f.net.irc.martyr.State;
import f00f.net.irc.martyr.Debug;
// InCommand is pending the next version of martyr.
// This Jife is written to Martyr v 0.3
//import f00f.net.irc.martyr.InCommand;
import f00f.net.irc.martyr.Command;
import f00f.net.irc.martyr.util.FullNick;
import f00f.net.irc.martyr.commands.MessageCommand;
import f00f.net.irc.martyr.commands.PartCommand;
import f00f.net.irc.martyr.commands.JoinCommand;
import f00f.net.irc.martyr.commands.QuitCommand;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.HashSet;
import java.util.StringTokenizer;
import java.util.Properties;
import java.util.Set;
import java.util.TreeSet;
import java.net.UnknownHostException;


// issues:
//
// The command structure is terrible and needs reworking.  It's far
// too complicated for what it does.  Each command defines a set of
// terms that should activate it, and that set gets discovered when
// the commands are added to the commandMap.
//
// The logging facility is in a dysfunctional limbo, and should either
// be made finished or eliminated. 
//
// The set command (to set per user variables) works fine, but since
// there isn't anything that uses it right now it is disabled.  The
// idea is to allow stuff like dictionary selection.
//

public class Jife implements Sender
{
	public static final String AUTO_JOIN_LIST = "jife.autoJoinList";
	public static final String PORT_PROPERTY = "jife.port";
	public static final String SERVER_PROPERTY = "jife.server";
	public static final String NICK_PROPERTY = "jife.nick";

	public static final String CHECKLOGGER_PROPERTY = "jife.checkLogger";
	public static final String SUGGESTIONSOURCE_PROPERTY = "jife.suggestionSource";
	public static final String DEFINITIONSOURCE_PROPERTY = "jife.definitionSource";
	
	// IRC Management
	private IRCConnection connection;
	private AutoReconnect autoReconnect;
	private String server;
	private int port;
	
	// Dictionary management
	private SuggestionSource suggestionEngine;

	// Logging
	private CheckLogger checkLogger;
	
	// Definition management
	private DefinitionSource lookupEngine;
	private DefinitionProxy lookupProxy;
	 
	// Settings
	private Properties config;
	private File configFile;
	// Maps a channel name to it's AutoJoin object.
	// Synchronize access!
	private Map autoJoinMap;

	// User management
	// A Map of FullNicks to Properties objects.
	private Map userProperties;
	
	// Maps command shortcuts to commands
	private Map commandMap;
	
	public Jife( Properties config, File configFile )
		throws Exception
	{
		this.config = config;
		this.configFile = configFile;

		commandMap = new HashMap();
		
		userProperties = Collections.synchronizedMap( new HashMap() );

		lookupProxy = new DefinitionProxy();
		

		// Suggestion source
		String sSource = config.getProperty( SUGGESTIONSOURCE_PROPERTY );
		if( sSource == null || sSource.equals("") )
		{
			Debug.println( this, SUGGESTIONSOURCE_PROPERTY + " not defined, no check command loaded.", Debug.VERBOSE );
		}
		else
		{
			suggestionEngine = (SuggestionSource)loadSource( sSource, config );

			addCommand( new JifeCheckCommand(), commandMap );
		}
		
		// Definition source
		String dSource = config.getProperty( DEFINITIONSOURCE_PROPERTY );
		if( dSource == null || dSource.equals("") )
		{
			Debug.println( this, DEFINITIONSOURCE_PROPERTY + " not defined, no define command loaded.", Debug.VERBOSE );
		}
		else
		{
			lookupEngine = (DefinitionSource)loadSource( dSource, config );
			lookupProxy.setSource( lookupEngine );

			addCommand( new JifeDefineCommand( lookupProxy, (Sender)this ), commandMap );
		}
		
		// Logger
		String checkLoggerName = config.getProperty( CHECKLOGGER_PROPERTY );
		if( checkLoggerName == null || checkLoggerName.equals(""))
		{
			Debug.println( this, CHECKLOGGER_PROPERTY + " not defined, no logger loaded.", Debug.VERBOSE );
			checkLogger = null;
		}
		else
		{
			checkLogger = loadCheckLogger( checkLoggerName, config );
		}

		//commandMap = setupCommands();

		// Some remaining commands
		addCommand( new HelpCommand(), commandMap );
		addCommand( new JifeJoinCommand(), commandMap );
		addCommand( new JifePartCommand(), commandMap );

		// Load IRC config
		loadConfig();
	}
	
	/**
	 * Loads a source of the given name from the given properties.
	 * The name.class is loaded and init is called on it.
	 * */
	public static Source loadSource( String name, Properties config )
		throws BadConfigException
	{
		try
		{
			String pName = name + ".class";
			String sourceClass = config.getProperty( pName );
			if( sourceClass == null )
			{
				throw new BadConfigException( pName  + " not defined.");
			}
			Source source =
				(Source)Class.forName( sourceClass ).newInstance();
			
			source.init( name, config );
			
			return source;
		}
		catch( BadConfigException bce )
		{
			throw bce;
		}
		catch( Exception e )
		{
			throw new BadConfigException( "" + e );
		}
	}

	public static CheckLogger loadCheckLogger( String name, Properties p )
		throws BadConfigException
	{
		try
		{
			String pName = name + ".class";
			String sourceClass = p.getProperty( pName );
			
			if( sourceClass == null )
			{
				throw new BadConfigException( pName  + " not defined.");
			}
			CheckLogger checkLogger =
				(CheckLogger)Class.forName( sourceClass ).newInstance();
			
			checkLogger.init( name, p );
			
			return checkLogger;
		}
		catch( BadConfigException bce )
		{
			throw bce;
		}
		catch( Exception e )
		{
			throw new BadConfigException( "" + e );
		}
	}
	
	/**
	 * For a given user, load the properties associated with that
	 * user.
	 * */
	public Properties getUserProperties( FullNick mask )
	{
		synchronized( userProperties )
		{
			Properties p = (Properties)userProperties.get( mask );
			
			if( p == null )
			{
				p = new Properties();
				userProperties.put( mask, p );
			}

			return p;
		}
	}
	
	//private Map setupCommands()
	//{
	//	Map m = new HashMap();
	//	
	//	//addCommand( new SetCommand( this ), m );
	//	
	//	return Collections.unmodifiableMap( m );
	//}

	private void addCommand( JifeCommand c, Map m )
	{
		Iterator cSet = c.getCommandSet().iterator();

		while( cSet.hasNext() )
		{
			m.put( cSet.next(), c );
		}
	}
	 
	private void loadConfig()
		throws BadConfigException
	{
		if( config.getProperty( PORT_PROPERTY ) == null )
			throw new BadConfigException( PORT_PROPERTY + " not defined." );
		if( config.getProperty( SERVER_PROPERTY ) == null )
			throw new BadConfigException( SERVER_PROPERTY + " not defined." );
		this.port = Integer.parseInt( config.getProperty( PORT_PROPERTY ) );
		this.server = config.getProperty( SERVER_PROPERTY );
		
		String nick = config.getProperty( NICK_PROPERTY );
		if( nick == null )
			nick = "jife";

		// Read the autoJoinSet
		autoJoinMap = new HashMap();
		String autoJoinsStr = config.getProperty( AUTO_JOIN_LIST );
		if( autoJoinsStr != null )
		{
			StringTokenizer autoJoinTokens = new StringTokenizer( autoJoinsStr, "," );
			while( autoJoinTokens.hasMoreTokens() )
			{
				autoJoinMap.put( autoJoinTokens.nextToken(), null );
			}
		}
		
		// Set up the connection
		connection = new IRCConnection();
		
		new AutoResponder( connection );
		new AutoReconnect( connection );
		new AutoRegister( connection, nick, nick, "Lexicographically Challenged" );
		new MessageMonitor( connection );
		
		loadAutoJoins();
	}

	private void loadAutoJoins()
	{
		Iterator autoJoins = autoJoinMap.keySet().iterator();
		while( autoJoins.hasNext() )
		{
			String chanName = (String)autoJoins.next();
			AutoJoin aj = new AutoJoin( connection, chanName );

			autoJoinMap.put( chanName, aj );
		}
	}

	private void saveConfig()
	{
		try
		{
			// First, load the existing config.  It may have changed on
			// disk since we last loaded it.
			config.load( new FileInputStream( configFile ));
			
			config.setProperty( PORT_PROPERTY, Integer.toString( this.port ) );
			config.setProperty( SERVER_PROPERTY, server );
			
			// Save the auto join list
			String totalList = "";
			Iterator autoJoins = autoJoinMap.keySet().iterator();
			while( autoJoins.hasNext() )
			{
				String chanName = (String)autoJoins.next();
				totalList += chanName;
				if( autoJoins.hasNext() )
				{
					totalList += ",";
				}
			}
	
			config.setProperty( AUTO_JOIN_LIST, totalList );
			
			config.store( new FileOutputStream( configFile ), "jife config" );
		}
		catch( IOException ioe )
		{
			Debug.println( this, "Can't save jife config: " + ioe, Debug.CRITICAL );
		}
	}
	
	public SuggestionSource getSuggestionSource()
	{
		return suggestionEngine;
	}

	public DefinitionSource getDefinitionSource()
	{
		return lookupEngine;
	}
	
	public void connect()
		throws UnknownHostException, IOException
	{
		connection.connect( server, port );
	}
	
	public void processMessage( MessageCommand msg )
	{

		boolean isToGroup = !msg.isPrivateToUs( connection.getClientState() );
		if( isToGroup )
		{
			String str = msg.getMessage();
			String ourNick = connection.getClientState().getNick().getNick();

			// Note: The string substring below only works because all of
			// these strings are the same length!
			String startStr1 = ourNick + ":";
			String startStr2 = ourNick + ";";
			String startStr3 = ourNick + " ";
			
			boolean nohelp = str.startsWith( startStr3 );
			
			// if the command is certainly to us, we want to help
			// if the command might not be to us (eg "jife is a nice
			// bot") then we don't want to help
			boolean helpOnError = !nohelp;

			if( str.startsWith( startStr1 )
				|| str.startsWith( startStr2 )
				|| str.startsWith( startStr3 ) )
			{
				// It's for us.
				
				// Old call:
				// processMessage( msg.getSource(), msg.getDest(), str.substring(startStr1.length()), true, nohelp);

				// New call: we use a thread
				Thread runner = new Thread(
					new RequestRunner( msg.getSource(), msg.getDest(),
						str.substring(startStr1.length()), isToGroup, helpOnError) );
				// false, true == not a group, do give help on error
				runner.start();
			}
			return;
		}
		
		Thread runner = new Thread( new RequestRunner( msg.getSource(), msg.getSource().getNick(), msg.getMessage(), isToGroup, true ) );
		// false, true == not a group, do give help on error
		runner.start();
	}
	
	private class RequestRunner implements Runnable
	{
		private FullNick sender;
		private String replyTo;
		private String input;
		private boolean toGroup;
		private boolean helpIfError;
		
		public RequestRunner( FullNick sender, String replyTo, String input, boolean toGroup, boolean helpIfError )
		{
			this.sender = sender;
			this.replyTo = replyTo;
			this.input = input;
			this.toGroup = toGroup;
			this.helpIfError = helpIfError;
		}

		public void run()
		{
			processMessage( sender, replyTo, input, toGroup, helpIfError );
		}
	}
	
	private void sendHelp( FullNick sender )
	{
		processMessage( sender, sender.getNick(), "help", false, false );
	}
	
	/*
	 * finds a command to run.  if a command is not found, sends help,
	 * unless helpIfError is false.
	 */
	private void processMessage( FullNick sender, String replyTo, String input, boolean toGroup, boolean helpIfError )
	{
		Debug.println( this, "Message to us from " + sender
			+ ", replying to " + replyTo + ": " + input, Debug.EXCESSIVE );
		StringTokenizer tokens = new StringTokenizer( input );
		
		String command = tokens.nextToken();
		
		try
		{
			JifeCommand jc = (JifeCommand)commandMap.get( command );
			if( jc != null )
			{
				Debug.println( this, "Found command.", Debug.EXCESSIVE );
				jc.processCommand( command, sender, replyTo, input, toGroup );
			}
			else if( command.equals( "quit" ) )
			{
				sendTo( replyTo, "Use the part command to get me to leave a channel." );
			}
			else
			{
				if( helpIfError )
					sendHelp( sender );
			}
		}
		catch( Throwable t )
		{
			t.printStackTrace( System.out );
			sendTo( replyTo, "Sorry, deterministic behaviour containment failed." );
			connection.sendCommand( new QuitCommand( "Mayday, mayday, mayday..." ) );
		}
	}
	
	public void sendTo( String dest, String msg )
	{
		connection.sendCommand( new MessageCommand( dest, msg ) );
	}
	
	public void sendTo( FullNick dest, String msg )
	{
		connection.sendCommand( new MessageCommand( dest, msg ) );
	}

	public void sendCommand( Command c )
	{
		connection.sendCommand( c );
	}
	
	public static void main( String args[] )
		throws Exception
	{
		if( args.length != 1 )
		{
			Debug.println( "Jife_init", "Must provide config filename.", Debug.CRITICAL );
			return;
		}
		File propsFile = new File( args[0] ).getAbsoluteFile();
		Properties props = new Properties();
		props.load( new FileInputStream( propsFile ));
		
		Jife jife = new Jife(props, propsFile);
		
		jife.connect();
	}

	public String toString()
	{
		return "Jife";
	}
	
	class MessageMonitor extends GenericAutoService
	{
		public MessageMonitor( IRCConnection c )
		{
			super( c );
			enable();
		} 
		
		protected void updateCommand( Command command )
		{
			if( command instanceof MessageCommand )
			{
				// A message!  How nice.
				processMessage( (MessageCommand)command );
			}
		}

		protected void updateState( State s )
		{
			// AutoReconnect now disables itself if it detects a QUIT,
			// so we don't have to do this anymore.
		
			//if( s == State.UNCONNECTED && shouldQuit )
			//{
			//	System.exit(0);
			//}
		}
	}
	
	private class JifeCheckCommand implements JifeCommand
	{
		public void processCommand( String commandStr, FullNick sender, String replyTo, String input, boolean toGroup )
			throws JifeException
		{
			StringTokenizer tokens = new StringTokenizer( input );
			tokens.nextToken(); // Throw away the "check"
			
			int numWords = tokens.countTokens();
			if( numWords == 0 )
			{
				sendHelp( sender.getNick(), "" );
				return;
			}
			if( numWords > 3 && toGroup )
			{
				sendTo( replyTo, "Sorry " + sender + ", you can't get more than 3 words checked in the channel at a time." );
				sendTo( sender, "Please be warned that most networks do not appreciate flood attempts." );	
				return;
			}

			while( tokens.hasMoreTokens() )
			{
				String word = tokens.nextToken();

				List results = suggestionEngine.getSuggestions( word );

				if( checkLogger != null )
				{
					checkLogger.logCheck( word, sender.getNick(), replyTo, results );
				}

				Iterator resultsI = results.iterator();
				String sendStr = word + ": ";

				if( !resultsI.hasNext() )
				{
					sendTo( replyTo, sendStr + "No results." );
				}
				else do
				{
					SuggestedWord suggestedWord =
						(SuggestedWord)resultsI.next();

					sendStr = sendStr + suggestedWord.getSuggested();

					if( ! resultsI.hasNext() )
					{
						sendTo( replyTo, sendStr + "." );
					}
					else if( sendStr.length() > 240 )
					{
						sendTo( replyTo, sendStr + ", " );
						sendStr = "";
					}
					else
					{
						sendStr = sendStr + ", ";
					}
				}
				while( resultsI.hasNext() );
			}
		}

		public void sendHelp( String replyTo, String input )
		{
			sendTo( replyTo, " \"check <word> <word> ... \" will cause jife to offer correct spellings of the words given." );
		}
		/**
		 * Returns a set of strings that represents the commands this
		 * command would like to respond to.
		 */
		public Set getCommandSet()
		{
			Set s = new HashSet();
			s.add( "check" );
			s.add( "c" );
			return s;
		}
		/**
		 * @return a short summary of what this command does.
		 */
		public String getSummary()
		{
			return "Directs jife to offer correct spellings.";
		}
	}

		private class JifePartCommand implements JifeCommand
	{
		public void processCommand( String commandStr, FullNick sender, String replyTo, String input, boolean toGroup )
			throws JifeException
		{
			StringTokenizer tokens = new StringTokenizer( input );
			if( tokens.countTokens() == 2 )
			{
				tokens.nextToken();
				String chanName = tokens.nextToken();

				synchronized( autoJoinMap )
				{
					AutoJoin aj = (AutoJoin)autoJoinMap.get( chanName );
					// Trying to part a channel that we aren't in?
					if( aj == null )
						return;
					aj.disable();
					connection.sendCommand( new PartCommand( chanName ) );
				
					autoJoinMap.remove( chanName );
					saveConfig();
				}
			}
			else
				sendHelp( sender.getNick(), input );
		}

		public void sendHelp( String replyTo, String input )
		{
			sendTo( replyTo, " \"part #channame\" will cause jife to leave a channel." );
		}
		/**
		 * Returns a set of strings that represents the commands this
		 * command would like to respond to.
		 */
		public Set getCommandSet()
		{
			return Collections.singleton( "part" );
		}
		/**
		 * @return a short summary of what this command does.
		 */
		public String getSummary()
		{
			return "Directs jife to leave a channel.";
		}
	}

	private class JifeJoinCommand implements JifeCommand
	{
		public void processCommand( String commandStr, FullNick sender, String replyTo, String input, boolean toGroup )
			throws JifeException
		{
			StringTokenizer tokens = new StringTokenizer( input );
			if( tokens.countTokens() == 2 )
			{
				tokens.nextToken(); // strip off "help"
				String chanName = tokens.nextToken();
				
				AutoJoin aj = new AutoJoin( connection, chanName );
				synchronized( autoJoinMap )
				{
					autoJoinMap.put( chanName, aj );
					saveConfig();
				}
			}
			else
				sendHelp( sender.getNick(), input );
		}

		public void sendHelp( String replyTo, String input )
		{
			sendTo( replyTo, " \"join #channame\" will cause jife to join and stay in a channel." );
		}
		/**
		 * Returns a set of strings that represents the commands this
		 * command would like to respond to.
		 */
		public Set getCommandSet()
		{
			return Collections.singleton( "join" );
		}
		/**
		 * @return a short summary of what this command does.
		 */
		public String getSummary()
		{
			return "Directs jife to join a channel.";
		}
	}

	/**
	 * The help command is somewhat special.
	 */
	private class HelpCommand implements JifeCommand
	{
		public void processCommand( String commandStr, FullNick sender, String replyTo, String input, boolean toGroup )
			throws JifeException
		{
			StringTokenizer tokens = new StringTokenizer( input );
			if( tokens.countTokens() == 2 )
			{
				tokens.nextToken(); // strip off "help"
				String cmdStr = tokens.nextToken();
				
				JifeCommand cmd = (JifeCommand)commandMap.get( cmdStr );
				if( cmd != null )
					cmd.sendHelp( sender.getNick(), input );
			}
			else
				sendHelp( sender.getNick(), input );
		}

		public void sendHelp( String replyTo, String input )
		{
			Iterator cmds = new TreeSet( commandMap.keySet() ).iterator();

			while( cmds.hasNext() )
			{
				String cmdStr = (String)cmds.next();
				JifeCommand cmd = (JifeCommand)commandMap.get( cmdStr );
				
				sendTo( replyTo, cmdStr + " - " + cmd.getSummary() );
			}
			
			sendTo( replyTo, " \"help <cmd>\" will get more help." );
		}
		/**
		 * Returns a set of strings that represents the commands this
		 * command would like to respond to.
		 */
		public Set getCommandSet()
		{
			return Collections.singleton( "help" );
		}
		/**
		 * @return a short summary of what this command does.
		 */
		public String getSummary()
		{
			return "Provides help.";
		}
	}
}

class DefinitionProxy implements DefinitionSource
{
	private DefinitionSource src = null;

	/**
	 * Does nothing.
	 * */
	public void init( String name, Properties p )
	{
	}

	public List getDefinition( String word ) throws JifeException
	{
		if( src == null )
		{
			throw new JifeException( "Definitions are not available.");
		}

		return src.getDefinition( word );
	}

	public void setSource( DefinitionSource src )
	{
		this.src = src;
	}
}

/**
 * Defines the define command, which looks up a word via the DICT
 * protocol.
 * */
// This command is special in the sense that it is not an
// inner class and so is a first stab at finding out what an abstract
// JifeCommand needs to live on its own.
class JifeDefineCommand implements JifeCommand
{
	private Sender sender;
	private DefinitionSource lookupEngine;
	public JifeDefineCommand( DefinitionSource source, Sender sender )
	{
		this.sender = sender;
		this.lookupEngine = source;
	}
	/**
	 * Sends a message. 
	 * */
	private void sendTo( String to, String msg )
	{
		sender.sendCommand( new MessageCommand( to, msg ) );
	}
	private void sendTo( FullNick to, String msg )
	{
		sender.sendCommand( new MessageCommand( to, msg ) );
	}
	public void processCommand( String commandStr, FullNick sender, String replyTo, String input, boolean toGroup )
		throws JifeException
	{
		StringTokenizer tokens = new StringTokenizer( input );
		tokens.nextToken();  // Throw out the command
		if( ! tokens.hasMoreTokens() )
		{
			sendTo( replyTo, "Must specify a word." );
			return;
		}
		String word = tokens.nextToken();
		if( lookupEngine != null )
		{
			sendTo( sender, "Processing request, may take a moment." );
			List result;
			try
			{
				result = lookupEngine.getDefinition( word );

				if( result == null )
				{
					sendTo( sender, "No results for \"" + word + "\".  (Misspelled?)");
					sendTo( replyTo, "No results for \"" + word + "\".");
					return;
				}
			}
			catch( Exception e )
			{
				sendTo( sender, "Error performing lookup. (" + e.getMessage() + ")" );
				return;
			}
			Iterator resultI = result.iterator();
			int groupCap = 4;
				int counter = 0;
				while( resultI.hasNext() )
				{

					if( counter++ == groupCap && toGroup )
					{
						while( resultI.hasNext() )
						{
							resultI.next();
							++counter;
						}
						sendTo( replyTo, "(" + groupCap + " out of " + counter + " lines sent to the channel, to prevent flooding)");
						return;
					}
					sendTo( replyTo, resultI.next().toString() );
					try
					{
						Thread.sleep(1000);
					}
					catch( Exception e )
					{

					}
				}
			}
			else
			{
				sendTo( replyTo, "Dictionary definition lookup is not configured." );
			}
		}

	public void sendHelp( String replyTo, String input )
	{
		sendTo( replyTo, " \"define <word>\" will cause jife to look up a definition." );
		sendTo( replyTo, "Set the define_dict variable to define which dictionary jife uses for your requests.<NYI>" );
	}
	/**
	 * Returns a set of strings that represents the commands this
	 * command would like to respond to.
	 */
	public Set getCommandSet()
	{
		Set s = new HashSet();
		s.add( "define" );
		s.add( "d" );
		return s;
	}
	/**
	 * @return a short summary of what this command does.
	 */
	public String getSummary()
	{
		return "Finds the dictionary definition of a word.";
	}
}


