package dammfine.tools;

import java.sql.*;
import java.io.*;
import java.util.Properties;

import dammfine.util.Metaphone;

/**
 * Prep takes a directory of scowl-style word lists and inserts them
 * into the jife database.
 * */
public class ScowlPrep
{
	// edit this
	public static final String DEF_db_username = "your_username_here";
	public static final String DEF_db_password = "your_password_here";

	// the JDBC driver for your database.
	public static final String DEF_db_driver = "org.postgresql.Driver";

	// the JDBC URL to the jife database.
	public static final String DEF_db_url = "jdbc:postgresql:jifedb";

	private int commitCounter;
	private PreparedStatement insertSmt;
	private Connection conn;
	private int wordCount;

	/**
	 * Prepares the data required to insert into the database.
	 * */
	private ScowlPrep()
		throws Exception
	{
		Properties p = new Properties();
		p.put("user", DEF_db_username );
		p.put("password", DEF_db_password );
		
		Driver d = (Driver)(Class.forName(DEF_db_driver).newInstance());
		
		conn = d.connect( DEF_db_url, p );
		 
		if(conn == null )
		{
			throw new IllegalStateException( "Null db conn!");
		}

		conn.setAutoCommit( false );

		insertSmt = conn.prepareStatement( 
			"insert into wordlist (word,metaphone,frequency) values "
			+ "(?,?,?)" );

		commitCounter = 0;
		wordCount = 0;
	}
	
	public void process( String word, int frequency )
		throws Exception
	{
		try
		{
			String metaphone ;
			try
			{
				metaphone = Metaphone.metaPhone( word );
			}
			catch( Exception e )
			{
				System.err.println("Error encoding: " + word );
				throw e;
			}

			
			insertSmt.setString( 1, word );
			insertSmt.setString( 2, metaphone ); 
			insertSmt.setShort( 3, (short)frequency );
			
			insertSmt.executeUpdate();

			++wordCount;
		}
		catch( Exception e )
		{
			System.err.println( "Error on word: " + word + ": " + frequency );
			throw e;
		}

		try
		{
			if( wordCount % 1000 == 0 )
				conn.commit();
		} 
		catch( Exception e )
		{
			System.err.println( "Commit error on " + word + ": " + frequency );
			throw e;
		}
	}
	
	public int getWordCount()
	{
		return wordCount;
	}
	
	public void finish()
		throws Exception
	{
		conn.commit();
	}
	
	public static void main( String[] args )
		throws Exception
	{
		// 10 20 35 50 60 65 70
		// read in a word
		// encode it
		// insert it into the dictionary.
		ScowlPrep prep = new ScowlPrep();

		doFrequency( prep, 10 );
		doFrequency( prep, 20 );
		doFrequency( prep, 35 );
		doFrequency( prep, 50 );
		doFrequency( prep, 60 );
		doFrequency( prep, 65 );
		doFrequency( prep, 70 );
	}

	public static void doFrequency( ScowlPrep prep, int frequency )
		throws Exception
	{
		System.out.println( "\nDoing: " +  frequency );
		BufferedReader reader ;
		try
		{
			reader = new BufferedReader( new FileReader( "words." + frequency ) );
		}
		catch( FileNotFoundException fnfe )
		{
			System.err.println( "Not found: " + fnfe );
			return;
		}

		loadFile( prep, frequency, reader );
	}

	public static void loadFile( ScowlPrep prep, int frequency, BufferedReader reader )
		throws Exception
	{
		String word ;
		int count = 0;
		long start = System.currentTimeMillis();

		try
		{
			while( (word = reader.readLine()) != null )
			{
				prep.process( word, frequency );
				++count;

				if(count % 100 == 0 && start + 1000 < System.currentTimeMillis() )
				{
					long now = System.currentTimeMillis();
					System.out.println( "Throughput: " 
						+ (count * 1000 / (double)(now-start) ) + " /s" );
				}
			}

			prep.finish();
		}
		finally
		{
			reader.close();
		}
	}
}








